/*
 * access.c -- the files with access control on open
 *
 * Copyright (C) 2001 Alessandro Rubini and Jonathan Corbet
 * Copyright (C) 2001 O'Reilly & Associates
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  The citation
 * should list that the code comes from the book "Linux Device
 * Drivers" by Alessandro Rubini and Jonathan Corbet, published
 * by O'Reilly & Associates.   No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 * $Id: access.c,v 1.21 2001/07/18 22:28:16 rubini Exp $
 */

 
#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#define __NO_VERSION__
#include <linux/module.h> /* get MOD_DEC_USE_COUNT, not the version string */
#include <linux/version.h> /* need it for conditionals in fortune.h */

#include <linux/kernel.h> /* printk() */
#include <linux/malloc.h> /* kmalloc() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/fcntl.h>
#include <linux/tty.h>    /* current->tty */

#include "fortune.h"        /* local definitions */
static int errno;

/*
 * These devices fall back on the main fortune operations. They only
 * differ in the implementation of open() and close()
 */

/*
 * The following deals with some of the 2.2 API changes.
 */
#ifdef LINUX_20

extern int fortune_lseek_20(struct inode *ino, struct file *f, off_t offset,
                int whence);
extern int fortune_read_20(struct inode *ino, struct file *f, char *buf,
                int count);
extern int fortune_write_20(struct inode *ino, struct file *f, const char *buf,
                int count);

#define fortune_llseek fortune_lseek_20
#define fortune_read   fortune_read_20
#define fortune_write  fortune_write_20
#endif



/************************************************************************
 *
 * The first device is the single-open one,
 *  it has an hw structure and an open count
 */

Fortune_Dev fortune_s_device;
int fortune_s_count = 0;
spinlock_t fortune_s_lock;

int fortune_s_open(struct inode *inode, struct file *filp)
{
    Fortune_Dev *dev = &fortune_s_device; /* device information */
    fortune_cookie *cookie = NULL;

    int num = NUM(inode->i_rdev);

    if (!filp->private_data && num > 0)
        return -ENODEV; /* not devfs: allow 1 device only */
    spin_lock(&fortune_s_lock);
    if (fortune_s_count) {
        spin_unlock(&fortune_s_lock);
        return -EBUSY; /* already open */
    }
    fortune_s_count++;
    spin_unlock(&fortune_s_lock);
    /* then, everything else is copied from the bare fortune device */

    if ( (filp->f_flags & O_ACCMODE) == O_WRONLY)
      /*fortune_trim(dev);*/

    cookie = (fortune_cookie *)filp->private_data;
    if (!cookie) {
	if ((filp->f_flags & O_ACCMODE) == O_WRONLY) {
	    if (!fortune_can_write_index ()) {
		return -EPERM;
	    }
	    cookie = fortune_new_fortune_cookie (dev, 0);
	} else {
	    if ((filp->f_flags & O_ACCMODE) == O_RDWR) {
		return -EPERM; /*read-write mode is nonsense. */
	    } else {
		/* O_RDONLY */
		cookie = fortune_new_fortune_cookie (dev, 1);
	    }
	}
	if (cookie->malloced == 0) {
	    kfree (cookie);
	    return -EIO; /* FIXME: too generic */
	}
        filp->private_data = cookie; /* for other methods */
    }
    MOD_INC_USE_COUNT;
    return 0;          /* success */
}

int fortune_s_release(struct inode *inode, struct file *filp)
{
    fortune_s_count--; /* release the device */
    MOD_DEC_USE_COUNT;
    return 0;
}

#ifdef LINUX_20

void fortune_s_release_20(struct inode *ino, struct file *f)
{
    fortune_s_release(ino, f);
}
#define fortune_s_release fortune_s_release_20
#define llseek lseek
#endif

/*
 * The other operations for the single-open device come from the bare device
 */
struct file_operations fortune_sngl_fops = {
    llseek:     fortune_llseek,
    read:       fortune_read,
    write:      fortune_write,
    open:       fortune_s_open,
    release:    fortune_s_release,
};


/************************************************************************
 *
 * Next, the "uid" device. It can be opened multiple times by the
 * same user, but access is denied to other users if the device is open
 */

Fortune_Dev fortune_u_device;
int fortune_u_count = 0;
uid_t fortune_u_owner = 0;
spinlock_t fortune_u_lock;

int fortune_u_open(struct inode *inode, struct file *filp)
{
    Fortune_Dev *dev = &fortune_u_device; /* device information */
    int num = NUM(inode->i_rdev);

    if (!filp->private_data && num > 0)
        return -ENODEV; /* not devfs: allow 1 device only */
    spin_lock(&fortune_u_lock);
    if (fortune_u_count && 
        (fortune_u_owner != current->uid) &&  /* allow user */
        (fortune_u_owner != current->euid) && /* allow whoever did su */
                    !capable(CAP_DAC_OVERRIDE)) { /* still allow root */
            spin_unlock(&fortune_u_lock);
            return -EBUSY;   /* -EPERM would confuse the user */
    }

    if (fortune_u_count == 0)
        fortune_u_owner = current->uid; /* grab it */

    fortune_u_count++;
    spin_unlock(&fortune_u_lock);

    /* then, everything else is copied from the bare fortune device */

    if ( (filp->f_flags & O_ACCMODE) == O_WRONLY)
      /*fortune_trim(dev);*/
    if (!filp->private_data)
        filp->private_data = fortune_new_fortune_cookie (dev, 1);
    MOD_INC_USE_COUNT;
    return 0;          /* success */
}

int fortune_u_release(struct inode *inode, struct file *filp)
{
    fortune_u_count--; /* nothing else */
    MOD_DEC_USE_COUNT;
    return 0;
}

#ifdef LINUX_20

void fortune_u_release_20(struct inode *ino, struct file *f)
{
    fortune_u_release(ino, f);
}
#define fortune_u_release fortune_u_release_20

#endif



/*
 * The other operations for the device come from the bare device
 */
struct file_operations fortune_user_fops = {
    llseek:     fortune_llseek,
    read:       fortune_read,
    write:      fortune_write,
    open:       fortune_u_open,
    release:    fortune_u_release,
};


/************************************************************************
 *
 * Next, the device with blocking-open based on uid
 */

Fortune_Dev fortune_w_device;
int fortune_w_count = 0;
uid_t fortune_w_owner = 0;
static DECLARE_WAIT_QUEUE_HEAD(fortune_w_wait);
spinlock_t fortune_w_lock;

int fortune_w_open(struct inode *inode, struct file *filp)
{
    Fortune_Dev *dev = &fortune_w_device; /* device information */
    int num = NUM(inode->i_rdev);

    if (!filp->private_data && num > 0)
        return -ENODEV; /* not devfs: allow 1 device only */
    spin_lock(&fortune_w_lock);
    while (fortune_w_count && 
      (fortune_w_owner != current->uid) &&  /* allow user */
      (fortune_w_owner != current->euid) && /* allow whoever did su */
      !capable(CAP_DAC_OVERRIDE)) {
        spin_unlock(&fortune_w_lock);
        if (filp->f_flags & O_NONBLOCK) return -EAGAIN; 
        interruptible_sleep_on(&fortune_w_wait);
        if (signal_pending(current)) /* a signal arrived */
          return -ERESTARTSYS; /* tell the fs layer to handle it */
        /* else, loop */
        spin_lock(&fortune_w_lock);
    }
    if (fortune_w_count == 0)
        fortune_w_owner = current->uid; /* grab it */
    fortune_w_count++;
    spin_unlock(&fortune_w_lock);

    /* then, everything else is copied from the bare fortune device */

    if ( (filp->f_flags & O_ACCMODE) == O_WRONLY)
      /*fortune_trim(dev);*/
    if (!filp->private_data)
        filp->private_data = fortune_new_fortune_cookie (dev, 1);
    MOD_INC_USE_COUNT;
    return 0;          /* success */
}

int fortune_w_release(struct inode *inode, struct file *filp)
{
    fortune_w_count--;
    if (fortune_w_count == 0)
        wake_up_interruptible(&fortune_w_wait); /* awake other uid's */
    MOD_DEC_USE_COUNT;
    return 0;
}


#ifdef LINUX_20

void fortune_w_release_20(struct inode *ino, struct file *f)
{
    fortune_w_release(ino, f);
}
#define fortune_w_release fortune_w_release_20

#endif



/*
 * The other operations for the device come from the bare device
 */
struct file_operations fortune_wusr_fops = {
    llseek:     fortune_llseek,
    read:       fortune_read,
    write:      fortune_write,
    open:       fortune_w_open,
    release:    fortune_w_release,
};

/************************************************************************
 *
 * Finally the `cloned' private device. This is trickier because it
 * involves list management, and dynamic allocation.
 */
devfs_handle_t fortune_priv_handle;    /* only used if devfs is there */

/* The clone-specific data structure includes a key field */
struct fortune_listitem {
    Fortune_Dev device;
    int key;
    struct fortune_listitem *next;
    
};

/* The list of devices, and a lock to protect it */
struct fortune_listitem *fortune_c_head;
spinlock_t fortune_c_lock;

/* Look for a device or create one if missing */
static Fortune_Dev *fortune_c_lookfor_device(int key)
{
    struct fortune_listitem *lptr, *prev = NULL;

    for (lptr = fortune_c_head; lptr && (lptr->key != key); lptr = lptr->next)
        prev=lptr;
    if (lptr) return &(lptr->device);

    /* not found */
    lptr = kmalloc(sizeof(struct fortune_listitem), GFP_ATOMIC);
    if (!lptr) return NULL;

    /* initialize the device */
    memset(lptr, 0, sizeof(struct fortune_listitem));
    lptr->key = key;
    /*fortune_trim(&(lptr->device));  initialize it */
    sema_init(&(lptr->device.sem), 1);

    /* place it in the list */
    if (prev)  prev->next = lptr;
    else       fortune_c_head = lptr;

    return &(lptr->device);
}

int fortune_c_open(struct inode *inode, struct file *filp)
{
    Fortune_Dev *dev;
    fortune_cookie *cookie = NULL;
    int key, num = NUM(inode->i_rdev);
 
    if (!filp->private_data && num > 0)
        return -ENODEV; /* not devfs: allow 1 device only */

    if (!current->tty) { 
        PDEBUG("Process \"%s\" has no ctl tty\n",current->comm);
        return -EINVAL;
    }
    key = MINOR(current->tty->device);

    /* look for a fortunec device in the list */
    spin_lock(&fortune_c_lock);
    dev = fortune_c_lookfor_device(key);
    spin_unlock(&fortune_c_lock);

    if (!dev) return -ENOMEM;

    /* then, everything else is copied from the bare fortune device */
    cookie = (fortune_cookie *)filp->private_data;
    if (!cookie) {
        if (num >= fortune_nr_devs) return -ENODEV;
	if ((filp->f_flags & O_ACCMODE) == O_WRONLY) {
	    if (!fortune_can_write_index ()) {
		return -EPERM;
	    }
	    cookie = fortune_new_fortune_cookie (fortune_devices + num, 0);
	} else {
	    if ((filp->f_flags & O_ACCMODE) == O_RDWR) {
		return -EPERM; /*read-write mode is nonsense. */
	    } else {
		/* O_RDONLY */
		cookie = fortune_new_fortune_cookie (fortune_devices + num, 1);
	    }
	}
	if (cookie->malloced == 0) {
	    kfree (cookie);
	    return -errno;
	}
        filp->private_data = cookie; /* for other methods */
    }
    
    MOD_INC_USE_COUNT;
    return 0;          /* success */
}

int fortune_c_release(struct inode *inode, struct file *filp)
{
    /*
     * Nothing to do, because the device is persistent.
     * A `real' cloned device should be freed on last close
     */
    MOD_DEC_USE_COUNT;
    return 0;
}


#ifdef LINUX_20

void fortune_c_release_20(struct inode *ino, struct file *f)
{
    fortune_c_release(ino, f);
}
#define fortune_c_release fortune_c_release_20

#endif



/*
 * The other operations for the device come from the bare device
 */
struct file_operations fortune_priv_fops = {
    llseek:   fortune_llseek,
    read:     fortune_read,
    write:    fortune_write,
    open:     fortune_c_open,
    release:  fortune_c_release,
};

/************************************************************************
 *
 * And the init and cleanup functions come last
 */

int fortune_access_init(void)
{

    /* Initialize spinlocks */
    spin_lock_init(&fortune_s_lock);
    spin_lock_init(&fortune_u_lock);
    spin_lock_init(&fortune_w_lock);
    spin_lock_init(&fortune_c_lock);

    /* and semaphores (used by read and write) */
    sema_init(&fortune_s_device.sem, 1);
    sema_init(&fortune_u_device.sem, 1);
    sema_init(&fortune_w_device.sem, 1);

    /* and file operations owners */
    SET_MODULE_OWNER(&fortune_sngl_fops);
    SET_MODULE_OWNER(&fortune_user_fops);
    SET_MODULE_OWNER(&fortune_wusr_fops);
    SET_MODULE_OWNER(&fortune_priv_fops);

#ifdef CONFIG_DEVFS_FS
    /* finally, create the devfs entry points */
    fortune_s_device.handle =
        devfs_register(fortune_devfs_dir, "single",
                       DEVFS_FL_AUTO_DEVNUM,
                       0, 0, S_IFCHR | S_IRUGO | S_IWUGO,
                       &fortune_sngl_fops,
                       &fortune_s_device);
    fortune_u_device.handle =
        devfs_register(fortune_devfs_dir, "user",
                       DEVFS_FL_AUTO_DEVNUM,
                       0, 0, S_IFCHR | S_IRUGO | S_IWUGO,
                       &fortune_user_fops,
                       &fortune_u_device);

    fortune_w_device.handle =
        devfs_register(fortune_devfs_dir, "wuser",
                       DEVFS_FL_AUTO_DEVNUM,
                       0, 0, S_IFCHR | S_IRUGO | S_IWUGO,
                       &fortune_wusr_fops,
                       &fortune_w_device);
    fortune_priv_handle =
        devfs_register(fortune_devfs_dir, "priv",
                       DEVFS_FL_AUTO_DEVNUM,
                       0, 0, S_IFCHR | S_IRUGO | S_IWUGO,
                       &fortune_priv_fops,
                       &fortune_priv_fops); /* any non-null value */
#endif    
    return 0;
}

/*
 * This is called by cleanup_module or on failure.
 * It is required to never fail, even if nothing was initialized first
 */
void fortune_access_cleanup(void)
{
    struct fortune_listitem *lptr, *prev;

    /*    fortune_trim(&fortune_s_device);  deallocate it 
	  fortune_trim(&fortune_u_device);
	  fortune_trim(&fortune_w_device);
	  */
    /* all the cloned devices */
    prev=NULL;
    for (lptr = fortune_c_head; lptr; lptr = lptr->next) {
      /*fortune_trim(&(lptr->device));*/
        if (prev) kfree(prev);
        prev=lptr;
    }
    if (prev) kfree(prev);
    fortune_c_head = NULL; /* overkill: we're unloading anyways */

    /* remove devfs entry points */
    devfs_unregister(fortune_s_device.handle);
    devfs_unregister(fortune_u_device.handle);
    devfs_unregister(fortune_w_device.handle);
    devfs_unregister(fortune_priv_handle);

    return;
}




