/*
 * fortune.h -- definitions for the char module
 *
 * Copyright (C) 2001 Alessandro Rubini and Jonathan Corbet
 * Copyright (C) 2001 O'Reilly & Associates
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  The citation
 * should list that the code comes from the book "Linux Device
 * Drivers" by Alessandro Rubini and Jonathan Corbet, published
 * by O'Reilly & Associates.   No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 * $Id: fortune.h,v 1.11 2001/07/18 22:28:18 rubini Exp $
 */

#ifndef _Makefile_H_
#define _Makefile_H_

#include <linux/ioctl.h> /* needed for the _IOW etc stuff used later */
/* version dependencies have been confined to a separate file */
#include "sysdep.h"

/*
 * Macros to help debugging
 */

#undef PDEBUG             /* undef it, just in case */
#ifdef Makefile_DEBUG
#  ifdef __KERNEL__
     /* This one if debugging is on, and kernel space */
#    define PDEBUG(fmt, args...) printk( KERN_DEBUG "fortune: " fmt, ## args)
#  else
     /* This one for user space */
#    define PDEBUG(fmt, args...) fprintf(stderr, fmt, ## args)
#  endif
#else
#  define PDEBUG(fmt, args...) /* not debugging: nothing */
#endif

#undef PDEBUGG
#define PDEBUGG(fmt, args...) /* nothing: it's a placeholder */

#ifndef Makefile_MAJOR
#define Makefile_MAJOR 0   /* dynamic major by default */
#endif

#ifndef Makefile_NR_DEVS
#define Makefile_NR_DEVS 4    /* fortune0 through fortune3 */
#endif

#ifndef Makefile_MAX_FORTUNE_SIZE 
#define Makefile_MAX_FORTUNE_SIZE 8096
#endif 

/*
 * The bare device is a variable-length region of memory.
 * Use a linked list of indirect blocks.
 *
 * "Fortune_Dev->data" points to an array of pointers, each
 * pointer refers to a memory area of Makefile_QUANTUM bytes.
 *
 * The array (quantum-set) is Makefile_QSET long.
 */
#ifndef Makefile_QUANTUM
#define Makefile_QUANTUM 4000
#endif

#ifndef Makefile_QSET
#define Makefile_QSET    1000
#endif

#ifndef Makefile_FILE
#define Makefile_FILE    "/home/novalis/fortune.txt"
#endif

#ifdef CONFIG_DEVFS_FS /* only if enabled, to avoid errors in 2.0 */
#include <linux/devfs_fs_kernel.h>
#else
  typedef void * devfs_handle_t;  /* avoid #ifdef inside the structure */
#endif

/*
 * This is somehow a hack: avoid ifdefs in the cleanup code by declaring
 * an empty procedure as a placeholder for devfs_unregister. This is
 * only done *unless* <linux/devfs_fs_kernel.h> was included, as that
 * header already implements placeholder for all the devfs functions
 */
/*............................................... degin-tag devfs-ifdef */
#ifndef DEVFS_FL_DEFAULT
extern inline void devfs_unregister(devfs_handle_t de) {}
#endif

extern devfs_handle_t fortune_devfs_dir;


typedef struct Fortune_Dev {
   devfs_handle_t handle;    /* only used if devfs is there */
   unsigned int access_key;  /* used by fortuneuid and fortunepriv */
   struct semaphore sem;     /* mutual exclusion semaphore     */
} Fortune_Dev;

typedef struct fortune_cookie {
  struct Fortune_Dev *fdev;
  char *ftext;
  int ftextsize;
  int malloced;
  int writing;
} fortune_cookie;

/*
 * Split minors in two parts
 */
#define TYPE(dev)   (MINOR(dev) >> 4)  /* high nibble */
#define NUM(dev)    (MINOR(dev) & 0xf) /* low  nibble */

/*
 * Different minors behave differently, so let's use multiple fops
 */

extern struct file_operations fortune_fops;        /* simplest: global */
extern struct file_operations fortune_priv_fops;   /* private region   */
extern struct file_operations fortune_sngl_fops;   /* single open      */
extern struct file_operations fortune_user_fops;   /* single process   */
extern struct file_operations fortune_wusr_fops;   /* single user      */

/*
 * The different configurable parameters
 */
extern int fortune_major;     /* main.c */
extern int fortune_nr_devs;
extern int fortune_max_size;
extern char * fortune_file;
extern char * index_file;
extern Fortune_Dev *fortune_devices;
/*
 * Prototypes for shared functions
 */

int     fortune_access_init(void);
void    fortune_access_cleanup(void);

ssize_t fortune_read (struct file *filp, char *buf, size_t count,
                    loff_t *f_pos);
ssize_t fortune_write (struct file *filp, const char *buf, size_t count,
                     loff_t *f_pos);
loff_t  fortune_llseek (struct file *filp, loff_t off, int whence);
fortune_cookie *fortune_new_fortune_cookie (Fortune_Dev *dev, int pickfortune);

int fortune_can_write_index ();


#endif /* _Makefile_H_ */
