/*
 * main.c -- the fortune device main module
 *
 * Copyright (C) 2001 David Turner <novalis@novalis.org>
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 *
 * This code is derived from code in the book "Linux Device
 * Drivers" by Alessandro Rubini and Jonathan Corbet, published
 * by O'Reilly & Associates.   No warranty is attached;
 * O'Reilly and Associates cannot take responsibility for errors 
 * or fitness for use.
 * Copyright (C) 2001 Alessandro Rubini and Jonathan Corbet
 * Copyright (C) 2001 O'Reilly & Associates
 *
 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#define __KERNEL_SYSCALLS__

#include <linux/config.h>
#include <linux/module.h>

#include <linux/kernel.h>   /* printk() */
#include <linux/malloc.h>   /* kmalloc() */
#include <linux/fs.h>       /* everything... */
#include <linux/file.h>     /* file stuff for stat */
#include <linux/stat.h>     /* stat () */
#include <linux/errno.h>    /* error codes */
#include <linux/types.h>    /* size_t */
#include <linux/fcntl.h>    /* O_ACCMODE */
#include <linux/unistd.h>
#include <linux/random.h>

#include <asm/system.h>     /* cli(), *_flags */

#include "fortune.h"          /* local definitions */
#include "index.h"          /* local definitions */

/*
 * I don't use static symbols here, because we export no symbols
 */

int fortune_major =   Makefile_MAJOR;
int fortune_nr_devs = Makefile_NR_DEVS;    /* number of bare fortune devices */
int fortune_max_size = Makefile_MAX_FORTUNE_SIZE;
char * fortune_file = Makefile_FILE;
char * index_file;

MODULE_PARM(fortune_major,"i");
MODULE_PARM(fortune_nr_devs,"i");
MODULE_PARM(fortune_max_size,"i");
MODULE_PARM(fortune_file,"s");

MODULE_PARM_DESC (fortune_major, "The major device number");
MODULE_PARM_DESC (fortune_nr_devs, "The number of fortune devices");
MODULE_PARM_DESC (fortune_max_size, "The max size of new fortunes");
MODULE_PARM_DESC (fortune_file, "The fortune filename");

MODULE_AUTHOR("David Turner and Alessandro Rubini");

Fortune_Dev *fortune_devices; /* allocated in fortune_init_module */


/*
 * Different minors behave differently, so let's use multiple fops
 */

struct file_operations *fortune_fop_array[]={
    &fortune_fops,      /* type 0 */
    &fortune_priv_fops, /* type 1 */
    &fortune_sngl_fops, /* type 2 */
    &fortune_user_fops, /* type 3 */
    &fortune_wusr_fops  /* type 4 */
};

#define Makefile_MAX_TYPE 5

static int errno;
static struct semaphore fortunefile_lock;

static char * get_random_fortune (int fd, int indexfd) {
    int num_fortunes, start, end, idx;
    char *fortune;  

    lseek (indexfd, 0, 0);

    /*get number of fortunes */
    read (indexfd, (void *) &num_fortunes, sizeof (int)); 
    if (num_fortunes == 0) {
	printk (KERN_WARNING "no fortunes!\n");
	fortune = kmalloc (1, GFP_KERNEL);
	if (fortune) {
	    *fortune = 0;
	    return fortune;
	} else {
	    printk (KERN_WARNING "can't make a null fortune!");
	    return NULL;
	}
    }
    printk(KERN_WARNING "how many?\n");

    /*pick one at random */
    get_random_bytes (&idx,sizeof (int));
    idx &= 0x7fffffffL; /* abs */
    idx %= num_fortunes;
    lseek (indexfd, INDEX_HEADER_END + sizeof (int) * idx, 0);
    printk(KERN_WARNING "yeah, that one: %i\n", idx);
    read (indexfd, (void *) &start, sizeof (int));
    printk(KERN_WARNING "starting at %i\n", start);
    read (indexfd, (void *) &end, sizeof (int));
    printk(KERN_WARNING "and ending at %i\n", end);
    end -= delim_size;
    
    /*read it*/

    lseek (fd, start, 0);
    printk(KERN_WARNING "seeking your fortune\n");
    fortune = kmalloc (end - start, GFP_KERNEL);
    printk (KERN_WARNING "and finding a space for it: %p\n", fortune);
    if (fortune == 0) {
	printk (KERN_WARNING "Couldn't malloc space for fortune (too big? %i)\n", end - start);
	fortune = kmalloc (1, GFP_KERNEL);
	if (fortune) {
	    *fortune = 0;
	    return fortune;
	} else {
	    printk (KERN_WARNING "can't make a null fortune!");
	    return NULL;
	}
    }
    read (fd, fortune, end - start);
    printk (KERN_WARNING "reading it\n");
    fortune [end - start] = 0;
    printk (KERN_WARNING "Finishing\n");
    return fortune;
}

static time_t get_index_mtime (int fd) {
    time_t mtime = 0;
    lseek (fd, sizeof (int), 0);
    read (fd, (void *) &mtime, sizeof (time_t));
    return mtime;
}

static time_t get_mtime (int fd) {
    time_t mtime = -1;
    struct file * f;
    int err = 0;

    f = fget (fd);
    if (f) {
	struct dentry * dentry = f->f_dentry;
	
	struct inode * inode = dentry->d_inode;
	if (inode->i_op && inode->i_op->revalidate)
	    err = inode->i_op->revalidate(dentry);
	if (!err)
	    mtime = inode->i_mtime;
	fput(f);
    }

    return mtime;
}

/*
 * Open and close
 */
fortune_cookie *fortune_new_fortune_cookie (Fortune_Dev *dev, int pickfortune) {

    fortune_cookie *cookie = kmalloc (sizeof (fortune_cookie), GFP_KERNEL);
    cookie->fdev = dev;
    cookie->writing = 0;

    if (pickfortune) {
	time_t mtime;
	int fd, indexfd;

	mm_segment_t fs = get_fs ();
	set_fs (get_ds ());

	cookie->malloced = 0;
    
	indexfd = open (index_file, O_RDONLY, 0);
	if (indexfd < 0) {
	    printk(KERN_WARNING "Unable to load index '%s' : %i.\n", index_file, errno);
	    return cookie;
	}

	fd = open (fortune_file, O_RDONLY, 0);
	if (fd < 0) {
	    close (indexfd);
	    printk(KERN_WARNING "Unable to load fortunes '%s' : %i.\n", fortune_file, errno);	    
	    return cookie;
	}

	/* Do we need to reindex? */
	mtime = get_mtime (fd);
	if (mtime == -1) return cookie;
	if (mtime > get_index_mtime (indexfd)) {
	    printk (KERN_WARNING "Needs reindexing!\n");
	    errno = -EIO;
	    close (fd);
	    close (indexfd);
	    return cookie;
	} else {
	    printk (KERN_WARNING "No reindexing for you!\n");
	}

	cookie->ftext = get_random_fortune (fd, indexfd);
	if (cookie->ftext) {
	    cookie->ftextsize = strlen (cookie->ftext);
	    cookie->malloced = cookie->ftextsize + 1;
	}
	close (fd);
	close (indexfd);

	set_fs (fs);

    } else {
	cookie->ftext = kmalloc (4096, GFP_KERNEL);
	cookie->malloced = 4096;
	cookie->ftext [0] = 0;
	cookie->ftextsize = 0;
    }
    return cookie;
}

/* blech.  There's gotta be a better way */
int fortune_can_write_index () {
    int indexfd = open (index_file, O_RDWR, 0);
    if (indexfd < 0) {
	return 0;
    }
    close (indexfd);
    return 0;
}

/* In fortune_open, the fop_array is used according to TYPE(dev) */
int fortune_open(struct inode *inode, struct file *filp)
{
    fortune_cookie *cookie; /* device information */
    int num = NUM(inode->i_rdev);
    int type = TYPE(inode->i_rdev);

    /*
     * the type and num values are only valid if we are not using devfs.
     * However, since we use them to retrieve the device pointer, we
     * don't need them with devfs as filp->private_data is already
     * initialized
     */

    /*
     * If private data is not valid, we are not using devfs
     * so use the type (from minor nr.) to select a new f_op
     */
    if (!filp->private_data && type) {
        if (type > Makefile_MAX_TYPE) return -ENODEV;
        filp->f_op = fortune_fop_array[type];
        return filp->f_op->open(inode, filp); /* dispatch to specific open */
    }

    /* type 0, check the device number (unless private_data valid) */
    cookie = (fortune_cookie *)filp->private_data;
    if (!cookie) {
        if (num >= fortune_nr_devs) return -ENODEV;
	if ((filp->f_flags & O_ACCMODE) == O_WRONLY) {
	    if (!fortune_can_write_index ()) {
		return -EPERM;
	    }
	    cookie = fortune_new_fortune_cookie (fortune_devices + num, 0);
	} else {
	    if ((filp->f_flags & O_ACCMODE) == O_RDWR) {
		return -EPERM; /*read-write mode is nonsense. */
	    } else {
		/* O_RDONLY */
		cookie = fortune_new_fortune_cookie (fortune_devices + num, 1);
	    }
	}
	if (cookie->malloced == 0) {
	    kfree (cookie);
	    return -EIO; /* FIXME: too generic */
	}
        filp->private_data = cookie; /* for other methods */
    }
    
    return 0;          /* success */
}

static void fortune_append (fortune_cookie *cookie) {

    int fd;
    int written, end, fortunes;
    struct timeval tv;
    mm_segment_t fs = get_fs ();

    set_fs (get_ds ());
	
    fd = open (fortune_file, O_RDWR, 0);

    if (fd == -1) {
	printk(KERN_WARNING "Unable to open '%s' : %i.\n", fortune_file, errno);
	return;
    }

    end = lseek (fd, 0, 2) + cookie->ftextsize + delim_size;
    written = write (fd, cookie->ftext, cookie->ftextsize);
    if (written != cookie->ftextsize) {
	printk(KERN_WARNING "Couldn't write all to '%s' : %i written, errno = %i.\n", fortune_file, written, errno);
    } else {
	printk(KERN_WARNING "Wrote all to '%s' : %i written, errno = %i.\n", fortune_file, written, errno);
    }
    write (fd, delimiter, delim_size);
    close (fd);


    /* Now update index */
    fd = open (index_file, O_RDWR, 0);
    if (fd == -1) {
	printk(KERN_WARNING "Unable to open '%s' : %i.\n", index_file, errno);
	return;
    }
    /* Write new fortune end index first because otherwise some other process could try for the last fortune before it's been written.*/
    lseek (fd, 0, 2);
    written = write (fd, (void *) &end, sizeof (int));
    if (written != sizeof (int)) {
	printk (KERN_WARNING "Couldn't write all to '%s' : %i written, errno = %i.\n", index_file, written, errno);
    } else {
	printk (KERN_WARNING "Wrote all to '%s' : %i written, errno = %i.\n", index_file, written, errno);
    }

    lseek (fd, 0, 0);
    /*read elems*/
    read (fd, (void *) &fortunes, sizeof (int));
    fortunes ++;
    printk(KERN_WARNING "Inc fortune_nr %i.\n", fortunes);
    lseek (fd, 0, 0);
    /*write elems, new mtime*/
    
    get_fast_time (&tv);
    write (fd, (void *) &fortunes, sizeof (int));
    write (fd, (void *) &tv.tv_sec, sizeof (time_t));
    close (fd);

    set_fs (fs);

    
}

int fortune_release(struct inode *inode, struct file *filp) {
    fortune_cookie *cookie = (fortune_cookie *) filp->private_data;
    if (cookie->writing) {
      if (down_interruptible (&fortunefile_lock)) 
	return -ERESTARTSYS;
      printk (KERN_WARNING "Start append\n");
      fortune_append (cookie);
      printk (KERN_WARNING "End append\n");
      up (&fortunefile_lock);
    }
    kfree (cookie->ftext);
    kfree (cookie);

    return 0;
}

/*
 * Data management: read and write
 */

ssize_t fortune_read(struct file *filp, char *buf, size_t count,
                loff_t *f_pos)
{
    fortune_cookie *cookie = (fortune_cookie *) filp->private_data;
    ssize_t ret = 0;

    if (down_interruptible(&cookie->fdev->sem))
      return -ERESTARTSYS;

    if (*f_pos >= cookie->ftextsize)
        goto out;

    if (*f_pos + count > cookie->ftextsize)
      count = cookie->ftextsize - *f_pos;

    if (copy_to_user (buf, cookie->ftext + *f_pos, count)) {
        ret = -EFAULT;
	goto out;
    }
    *f_pos += count;
    ret = count;

 out:
    up(&cookie->fdev->sem);
    return ret;

}

ssize_t fortune_write(struct file *filp, const char *buf, size_t count,
                loff_t *f_pos)
{

    fortune_cookie * cookie = (fortune_cookie *) filp->private_data;
    ssize_t ret = -ENOMEM;
    char *newmem;

    if (down_interruptible(&cookie->fdev->sem))
      return -ERESTARTSYS;

    cookie->writing = 1;
    
    /*not enough space? malloc more! If we're at the end of our tether already, tell the writer that we're out of space */
    if (*f_pos == fortune_max_size && count) {
	return -ENOSPC;
    }
    if (*f_pos + count >= cookie->malloced) {
      if (*f_pos + count > fortune_max_size)
	count = fortune_max_size - *f_pos;

      newmem = kmalloc (*f_pos + count + 1, GFP_KERNEL);
      memcpy (newmem, cookie->ftext, cookie->ftextsize + 1);
      kfree (cookie->ftext);
      cookie->ftext = newmem;

      cookie->malloced = *f_pos + count + 1;
    }

    if (cookie->ftextsize < *f_pos + count) {
      cookie->ftextsize = *f_pos + count;
    }

    cookie->ftext [cookie->ftextsize] = 0;

    copy_from_user (cookie->ftext + *f_pos, buf, count);
    
    *f_pos += count;
    ret = count;

    up(&cookie->fdev->sem);
    return ret;
}

/*
 * The "extended" operations -- only seek
 */

loff_t fortune_llseek(struct file *filp, loff_t off, int whence)
{
    fortune_cookie *cookie = (fortune_cookie *) filp->private_data;
    loff_t newpos;

    switch(whence) {
      case 0: /* SEEK_SET */
        newpos = off;
        break;

      case 1: /* SEEK_CUR */
        newpos = filp->f_pos + off;
        break;

      case 2: /* SEEK_END */
        newpos = cookie->ftextsize + off;
        break;

      default: /* can't happen */
        return -EINVAL;
    }
    if (newpos < 0) return -EINVAL;
    filp->f_pos = newpos;
    return newpos;
}


/*
 * The following wrappers are meant to make things work with 2.0 kernels
 */
#ifdef LINUX_20
int fortune_lseek_20(struct inode *ino, struct file *f,
                off_t offset, int whence)
{
    return (int)fortune_llseek(f, offset, whence);
}

int fortune_read_20(struct inode *ino, struct file *f, char *buf, int count)
{
    return (int)fortune_read(f, buf, count, &f->f_pos);
}

int fortune_write_20(struct inode *ino, struct file *f, const char *b, int c)
{
    return (int)fortune_write(f, b, c, &f->f_pos);
}

void fortune_release_20(struct inode *ino, struct file *f)
{
    fortune_release(ino, f);
}

/* Redefine "real" names to the 2.0 ones */
#define fortune_llseek fortune_lseek_20
#define fortune_read fortune_read_20
#define fortune_write fortune_write_20
#define fortune_release fortune_release_20
#define llseek lseek
#endif  /* LINUX_20 */

struct file_operations fortune_fops = {
    llseek:     fortune_llseek,
    read:       fortune_read,
    write:      fortune_write,
    open:       fortune_open,
    release:    fortune_release,
};

/*
 * Finally, the module stuff
 */

#ifdef CONFIG_DEVFS_FS
devfs_handle_t fortune_devfs_dir;
static char devname[4];
#endif

/*
 * The cleanup function is used to handle initialization failures as well.
 * Thefore, it must be careful to work correctly even if some of the items
 * have not been initialized
 */
void fortune_cleanup_module(void)
{
    int i;

#ifndef CONFIG_DEVFS_FS
    /* cleanup_module is never called if registering failed */
    unregister_chrdev(fortune_major, "fortune");
#endif

#ifdef Makefile_DEBUG /* use proc only if debugging */
    fortune_remove_proc();
#endif
    if (fortune_devices) {
        for (i=0; i<fortune_nr_devs; i++) {
            /* the following line is only used for devfs */
            devfs_unregister(fortune_devices[i].handle);
        }
        kfree(fortune_devices);
    }

    /* and call the cleanup functions for friend devices */
    fortune_access_cleanup();

    /* once again, only for devfs */
    devfs_unregister(fortune_devfs_dir);
    kfree (index_file);
}


int fortune_init_module(void)
{
    int result, i, fnlen;

    SET_MODULE_OWNER(&fortune_fops);

#ifdef CONFIG_DEVFS_FS
    /* If we have devfs, create /dev/fortune to put files in there */
    fortune_devfs_dir = devfs_mk_dir(NULL, "fortune", NULL);
    if (!fortune_devfs_dir) return -EBUSY; /* problem */

#else /* no devfs, do it the "classic" way  */    

    /*
     * Register your major, and accept a dynamic number. This is the
     * first thing to do, in order to avoid releasing other module's
     * fops in fortune_cleanup_module()
     */
    result = register_chrdev(fortune_major, "fortune", &fortune_fops);
    if (result < 0) {
        printk(KERN_WARNING "fortune: can't get major %d\n",fortune_major);
        return result;
    }
    if (fortune_major == 0) fortune_major = result; /* dynamic */

#endif /* CONFIG_DEVFS_FS */
    /* 
     * allocate the devices -- we can't have them static, as the number
     * can be specified at load time
     */
    fortune_devices = kmalloc(fortune_nr_devs * sizeof(Fortune_Dev), GFP_KERNEL);
    if (!fortune_devices) {
        result = -ENOMEM;
        goto fail;
    }
    memset(fortune_devices, 0, fortune_nr_devs * sizeof(Fortune_Dev));
    for (i=0; i < fortune_nr_devs; i++) {
        sema_init(&fortune_devices[i].sem, 1);
#ifdef CONFIG_DEVFS_FS
        sprintf(devname, "%i", i);
        devfs_register(fortune_devfs_dir, devname,
                       DEVFS_FL_AUTO_DEVNUM,
                       0, 0, S_IFCHR | S_IRUGO | S_IWUGO,
                       &fortune_fops,
                       fortune_devices+i);
#endif  
    }

    /* At this point call the init function for any friend device */
    if ( (result = fortune_access_init()) )
        goto fail;
    /* ... */

#ifndef Makefile_DEBUG
    EXPORT_NO_SYMBOLS; /* otherwise, leave global symbols visible */
#endif

#ifdef Makefile_DEBUG /* only when debugging */
    fortune_create_proc();
#endif

    sema_init (&fortunefile_lock, 1);
    fnlen = strlen (fortune_file);
    index_file = kmalloc (fnlen + 7, GFP_KERNEL);
    strcpy (index_file, fortune_file);
    strcpy (index_file + fnlen, ".index");

    printk (KERN_WARNING "Installed: %s %s\n", fortune_file, index_file);
    return 0; /* succeed */

  fail:
    printk (KERN_WARNING "fortune: failed for unspecified reason\n");
    fortune_cleanup_module();
    return result;
}



module_init(fortune_init_module);
module_exit(fortune_cleanup_module);
